import logging
from pathlib import Path
from typing import Any, cast

from aiohttp import web
from aiohttp.web_app import Application
from raphson_music_client import RaphsonMusicClient

from raphson_music_headless.player.mpv_player import MPVAudioPlayer
from raphson_music_headless.player.vlc_player import VLCAudioPlayer

from .config import Config
from .downloader import Downloader
from .player import AudioPlayer

_LOGGER = logging.getLogger(__name__)


class App:
    config: Config
    client: RaphsonMusicClient
    player: AudioPlayer
    downloader: Downloader

    def __init__(self, config: Config):
        self.config = config
        self.client = RaphsonMusicClient()
        self.downloader = Downloader(self.client, config)

        if config.player == 'mpv':
            self.player = MPVAudioPlayer(self.client, self.downloader, config)
        elif config.player == 'vlc':
            self.player = VLCAudioPlayer(self.client, self.downloader, config)
        else:
            raise ValueError('invalid player: ', self.player)

    async def _root(self, _request: web.Request) -> web.StreamResponse:
        return web.FileResponse(Path(__file__).parent / "index.html")

    async def _state(self, _request: web.Request) -> web.StreamResponse:
        data: dict[str, Any] = {
            "playlists": {
                "all": list(self.downloader.all_playlists.keys()),
                "enabled": self.downloader.enabled_playlists,
            },
            "player": {
                "has_media": self.player.has_media(),
                "is_playing": self.player.is_playing(),
                "position": self.player.position(),
                "duration": self.player.duration(),
                "volume": self.player.get_volume(),
            },
        }
        if self.player.currently_playing and self.player.currently_playing.track:
            track = self.player.currently_playing.track
            data["currently_playing"] = {
                "path": track.path,
                "duration": track.duration,
                "title": track.title,
                "album": track.album,
                "album_artist": track.album_artist,
                "year": track.year,
                "artists": track.artists,
            }
        else:
            data["currently_playing"] = None

        return web.json_response(data)

    async def _image(self, _request: web.Request) -> web.StreamResponse:
        if self.player.currently_playing:
            return web.Response(
                body=self.player.currently_playing.image,
                status=200,
                content_type="image/webp",
            )

        return web.Response(status=404)

    async def _lyrics(self, _request: web.Request) -> web.StreamResponse:
        if self.player.currently_playing:
            lyrics = self.player.currently_playing.lyrics_text
            if lyrics:
                return web.Response(
                    body=self.player.currently_playing.lyrics_text,
                    status=200,
                    content_type="text/plain",
                )

        return web.Response(status=204)

    async def _list_tracks(self, request: web.Request) -> web.StreamResponse:
        playlist = request.query["playlist"]
        return web.Response(
            body=await self.client.list_tracks_response(playlist),
            status=200,
            content_type="application/json",
        )

    async def _stop(self, _request: web.Request) -> web.StreamResponse:
        self.player.stop()
        return web.Response(status=204)

    async def _pause(self, _request: web.Request) -> web.StreamResponse:
        self.player.pause()
        return web.Response(status=204)

    async def _play(self, _request: web.Request) -> web.StreamResponse:
        await self.player.play()
        return web.Response(status=204)

    async def _next(self, _request: web.Request) -> web.StreamResponse:
        await self.player.next(retry=False)
        return web.Response(status=204)

    async def _play_news(self, _request: web.Request) -> web.StreamResponse:
        await self.player.next(retry=False, force_news=True)
        return web.Response(status=204)

    async def _seek(self, request: web.Request) -> web.StreamResponse:
        position = int(await request.text())
        self.player.seek(position)
        return web.Response(status=204)

    async def _volume(self, request: web.Request) -> web.StreamResponse:
        volume = int(await request.text())
        self.player.set_volume(volume)
        return web.Response(status=204)

    async def _playlists(self, request: web.Request) -> web.StreamResponse:
        playlists = cast(list[str], await request.json())
        assert isinstance(playlists, list)
        for playlist in playlists:
            assert isinstance(playlist, str)
            assert playlist in self.downloader.all_playlists
        _LOGGER.info("Changed enabled playlists: %s", playlists)
        self.downloader.enabled_playlists = playlists
        return web.Response(status=204)

    async def _enqueue(self, request: web.Request) -> web.StreamResponse:
        await self.downloader.enqueue(await request.text())
        return web.Response(status=204)

    async def _play_track(self, request: web.Request) -> web.StreamResponse:
        await self.downloader.enqueue(await request.text(), front=True)
        await self.player.next(retry=False)
        return web.Response(status=204)

    async def setup(self, _app: Application):
        await self.client.setup(
            base_url=self.config.server,
            user_agent="Raphson-Music-Headless",
            token=self.config.token,
        )
        await self.downloader.setup()
        await self.player.setup()

    async def cleanup(self, _app: Application) -> None:
        _LOGGER.info("Stopping")
        await self.client.close()
        self.player.stop()

    def start(self, config: Config):
        routes = [
            web.get("/", self._root),
            web.get("/state", self._state),
            web.get("/image", self._image),
            web.get("/lyrics", self._lyrics),
            web.get("/list_tracks", self._list_tracks),
            web.post("/stop", self._stop),
            web.post("/pause", self._pause),
            web.post("/play", self._play),
            web.post("/next", self._next),
            web.post("/play_news", self._play_news),
            web.post("/seek", self._seek),
            web.post("/volume", self._volume),
            web.post("/playlists", self._playlists),
            web.post("/enqueue", self._enqueue),
            web.post("/play_track", self._play_track),
        ]

        app = web.Application()
        app.add_routes(routes)
        app.on_startup.append(self.setup)
        app.on_cleanup.append(self.cleanup)

        web.run_app(app, host=config.host, port=config.port)  # pyright: ignore[reportUnknownMemberType]
